//
// Syd: rock-solid application kernel
// src/syd-sys.rs: Calculate the memory usage of a given process or the parent process.
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0

use std::process::ExitCode;

use getargs::{Opt, Options};
use nix::{errno::Errno, libc::pid_t, unistd::Pid};
use procfs::process::MMapPath;
use syd::{
    err::SydResult,
    human_size,
    proc::{proc_smaps, proc_statm},
};

fn main() -> SydResult<ExitCode> {
    syd::set_sigpipe_dfl()?;

    // Configure syd::proc.
    syd::config::proc_init()?;

    let args = std::env::args().skip(1).collect::<Vec<_>>();
    let args: &'static Vec<String> = Box::leak(Box::new(args));
    let mut opts = Options::new(args.iter().map(String::as_str));
    let mut human = false; // -H
    let mut is_vm = false; // -V

    while let Some(opt) = opts.next_opt()? {
        match opt {
            Opt::Short('h') => {
                help();
                return Ok(ExitCode::SUCCESS);
            }
            Opt::Short('H') => {
                human = true;
            }
            Opt::Short('V') => {
                is_vm = true;
            }
            Opt::Short(opt) => {
                eprintln!("Unknown option: -{opt}!");
                help();
                return Ok(ExitCode::FAILURE);
            }
            Opt::Long(opt) => {
                eprintln!("Unknown option: --{opt}!");
                help();
                return Ok(ExitCode::FAILURE);
            }
        }
    }

    let pid = match opts.positionals().next() {
        None => Pid::parent().as_raw(),
        Some(pid) => match pid.parse::<pid_t>() {
            Ok(pid) => pid,
            Err(error) => {
                eprintln!("syd-mem: {error}");
                help();
                return Ok(ExitCode::FAILURE);
            }
        },
    };

    let size = if is_vm {
        match proc_statm(Pid::from_raw(pid)) {
            Ok(statm) => statm.size.saturating_mul(*syd::config::PAGE_SIZE),
            Err(error) => {
                eprintln!("syd-mem: {error}");
                return Ok(ExitCode::FAILURE);
            }
        }
    } else {
        match proc_mem(Pid::from_raw(pid)) {
            Ok(size) => size,
            Err(error) => {
                eprintln!("syd-mem: {error}");
                return Ok(ExitCode::FAILURE);
            }
        }
    };

    if human {
        println!("{}", human_size(size as usize));
    } else {
        println!("{size}");
    }

    Ok(ExitCode::SUCCESS)
}

fn help() {
    println!("Usage: syd-mem [-HV] [pid]");
    println!("Calculate the memory usage of a given process or the parent process and exit.");
    println!("-H    Print human-formatted size");
    println!("-V    Print virtual memory size");
}

/// Calculates process memory usage.
///
/// This function uses the `procfs` crate to obtain detailed memory maps
/// from `/proc/[pid]/smaps`. It sums multiple memory usage values reported in these maps
/// to calculate a more comprehensive total memory usage.
///
/// # Arguments
///
/// * `process` - `Process` instance representing the process.
///
/// # Returns
///
/// This function returns a `Result<u64, Errno>`.
///
/// # Errors
///
/// This function returns an error if it fails to retrieve the process's memory maps,
/// typically due to insufficient permissions or an invalid process ID.
fn proc_mem(pid: Pid) -> Result<u64, Errno> {
    proc_smaps(pid).map(|maps| {
        let mut total_size: u64 = 0;
        for map in &maps {
            match &map.0.pathname {
                MMapPath::Path(_) | MMapPath::Anonymous | MMapPath::Stack | MMapPath::Other(_) => {
                    let pss = map.0.extension.map.get("Pss").copied().unwrap_or(0);
                    let private_dirty = map
                        .0
                        .extension
                        .map
                        .get("Private_Dirty")
                        .copied()
                        .unwrap_or(0);
                    let shared_dirty = map
                        .0
                        .extension
                        .map
                        .get("Shared_Dirty")
                        .copied()
                        .unwrap_or(0);

                    total_size = total_size.saturating_add(
                        pss.saturating_add(private_dirty)
                            .saturating_add(shared_dirty),
                    );
                }
                _ => (),
            }
        }

        total_size
    })
}
